/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Classifier.java
 *    Copyright (C) 1999 Eibe Frank, Len Trigg
 *
 */

package weka.classifiers;

import java.util.*;

import weka.core.*;
import weka.roughset.*;


public abstract class RuleClassifier extends Classifier 
  implements OptionHandler {

  protected Rules rules = null;
  protected String ruleStrengthOption = "s";
  protected double accuracy = 1;
  protected double reducedSetAccuracy = 1;
  

  /**
     All get-set and TipText-methods are necessary in order to use
     this class in the Explorer GUI and be able to set the class
     fields from there.
  */

  /**
   * Get the value of ruleStrengthOption.
   * @return value of ruleStrengthOption.
   */
  public String getRuleStrengthOption() {
    return ruleStrengthOption;
  }
  
  /**
   * Set the value of ruleStrengthOption.
   * @param v  Value to assign to ruleStrengthOption.
   */
  public void setRuleStrengthOption(String  v) {
    this.ruleStrengthOption = v;
  }
  
  public String ruleStrengthOptionTipText(){
    return "Set rule strength option: {s|m}";
  }

  /**
   * Get the value of reducedSetAccuracy.
   * @return value of reducedSetAccuracy.
   */
  public double getReducedSetAccuracy() {
    return reducedSetAccuracy;
  }
  
  /**
   * Set the value of reducedSetAccuracy.
   * @param v  Value to assign to reducedSetAccuracy.
   */
  public void setReducedSetAccuracy(double  v) {
    this.reducedSetAccuracy = v;
  }

  public String reducedSetAccuracyTipText() {
    return 
      "Reduce the rule sets as long as the " 
      + "accuracy is at least this value";
  }
  
  /**
   * Get the value of accuracy.
   * @return value of accuracy.
   */
  public double getAccuracy() {
    return accuracy;
  }
  
  /**
   * Set the value of accuracy.
   * @param v  Value to assign to accuracy.
   */
  public void setAccuracy(double  v) {
    this.accuracy = v;
  }
  
  public String accuracyTipText() {
    return "Reduce the rules until the accuracy is at least this value";
  }

  public Enumeration listOptions() {

    Vector l = new Vector();

    l.addElement(new Option("{s|m|r|d|p|c} Rule strength calculation.\n\n"
			    + "\ts = Simple Strength (default)\n"
			    + "\tm = Maximal Strength\n"
			    + "\tr = Global Strength (not tested)\n"
			    + "\td = Stability Strength\n"
			    + "\tp = Global Stability Strength (not tested)\n"
			    + "\tc = Conditional Probability Strength\n"
			    + "\t\t(the only rule strength in the\n"
			    + "\t\t original LERS) (not tested)\n"
			    + "\tu = Support Strength (not tested)\n"
			    + "\ta = Partial support strength (not tested)",
			    "S", 1, "-S <option>"));
    l.addElement(new Option("\tApproximate rules." 
			    + " Provide value between 0 and 1",
			    "A", 1, "-A <val>"));
    l.addElement(new Option("\tReduce rule sets." 
			    + " Provide value between 0 and 1",
			    "R", 1, "-R <val>"));
    return l.elements();

  }

  /**
   * Parses a given list of options.
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   */

  public void setOptions(String[] options) throws Exception{    

    String ruleStrength = Utils.getOption('S', options);
    if (ruleStrength.length() > 0) {
      ruleStrengthOption = ruleStrength;
    } 

    String approximationOption = Utils.getOption('A', options);
    if ( approximationOption.length() > 0) {
      accuracy = Double.parseDouble(approximationOption);
      if (accuracy  < 0 || accuracy > 1 ) {
	throw new Exception("Illegal approximation threshold. "
			    + "Choose a value between 0 and 1");
      } // end of if ()
    } // end of if ()

    String reductionOption = Utils.getOption('R', options);
    if ( reductionOption.length() > 0) {
      reducedSetAccuracy = Double.parseDouble(reductionOption);
      if (reducedSetAccuracy  < 0 || reducedSetAccuracy > 1 ) {
	throw new Exception("Illegal reduction threshold. "
			    + "Choose a value between 0 and 1");
      } // end of if ()
    } // end of if ()
    if ( reducedSetAccuracy < 1 && accuracy < 1 ) {
      throw new Exception("Illegal combination of " 
			  + "rule approximation options.");
    } // end of if ()
  }

  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] options = new String [10];
    int current = 0;

    if ( !ruleStrengthOption.equals(new String("s")) ) {
      options[current++] = "-S"; options[current++] = ruleStrengthOption;
    } // end of if ()

    if ( !(accuracy == 1) ) {
      options[current++] = "-A"; options[current++] = "" + accuracy;
    } // end of if ()

    if ( !(reducedSetAccuracy == 1) ) {
      options[current++] = "-R"; 
      options[current++] = "" + reducedSetAccuracy;
    } // end of if ()

    /* Fill out the rest */

    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }
  
  /**
   * Classifies a given instance. 
   *
   * @param instance the instance to be classified
   * @return index of the predicted class as a double
   * if the class is nominal, otherwise the predicted value
   * @exception Exception if instance could not be classified
   * successfully
   */
  public double classifyInstance(Instance instance) throws Exception {
    return rules.getDecision(instance,ruleStrengthOption);
  } 


}
