/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * CutPoint
 * Copyleft (C) 2002 Ola Leifler
 *
 */


package weka.roughset;

import java.io.*;
import java.util.*;
import weka.core.*;
import weka.filters.*; 

public class CutPoint extends Object {
  private Attribute attr;
  private double cut;
  private double classValue;
  private Instances insts;
  private int discerned;
  private int left;
  private int right;
  private int[] leftModClass;
  private int[] rightModClass;
  private int debug = 0;

  /* Update discernibility values for the first (virtual) cut on each attribute */
  
  private void updateValues() {
    leftModClass = new int[insts.numClasses()];
    rightModClass = new int[insts.numClasses()];
    Arrays.fill(leftModClass,0);
    Arrays.fill(rightModClass,0);
    left = 1;
    right = insts.numInstances()-1;
    leftModClass[(int)getClassValue()] = 1;

    /**
       All instances are said to be to the right of this cut...
     */

    for (int i=0;i<insts.numInstances();i++) {
      rightModClass[(int)insts.instance(i).classValue()] ++;
    } // end of for ()

    /**
       ...except for one
     */

    rightModClass[(int)getClassValue()]--;
    discerned = right - Utils.sum(rightModClass);
  }

  /* Update discernibility values for a subsequent cut in the dataset */

  private void updateValues(CutPoint c) { 
    left = c.getLeftOfCut()+1;
    right = c.getRightOfCut()-1;
    leftModClass = new int[insts.numClasses()];
    rightModClass = new int[insts.numClasses()];
    System.arraycopy(c.getLeftModClass(),0,leftModClass,0,insts.numClasses());
    System.arraycopy(c.getRightModClass(),0,rightModClass,0,insts.numClasses());
    leftModClass[(int)getClassValue()]++;
    rightModClass[(int)getClassValue()]--;
    discerned = c.getDiscerned() 
      + (c.getRightOfCut() - c.getLeftOfCut()) 
      - (c.getRightModDecision() - c.getLeftModDecision());
  }

  /* Constructor for the first cutpoint of attribute a */

  public CutPoint(Attribute a, double c, Collection inBetween, Instances i) 
    throws UnsupportedAttributeTypeException {
    if ( !i.classAttribute().isNominal() ) {
      throw(new 
	UnsupportedAttributeTypeException("Can only create cutpoints"
					  + " for nominal classes, "
					  + i.classAttribute().name()
					  + " is of type "
					  + (i.classAttribute().
					     isNumeric() ? 
					     "numeric" : 
					     "string")));
    } // end of if ()
    Iterator iter = inBetween.iterator();
    CutPoint virtualCut = new CutPoint(a,c,((Instance)iter.next()).classValue(),i);
    while ( iter.hasNext() ) {
      virtualCut = new CutPoint(virtualCut,c,((Instance)iter.next()).classValue());
    } // end of while ()
    this.copyValues(virtualCut);
    if ( debug > 0 ) {
      System.out.println("************ <1> ************");
      System.out.println("First cut :\n" + this.toInfoString());      
      System.out.println("*****************************");
    } // end of if ()
  }

  private CutPoint(Attribute a,double val,double classVal,Instances i){
    attr = a;
    cut = val;
    classValue = classVal;
    insts = i;
    updateValues();
  }

  public CutPoint(CutPoint c, double cutValue, Collection inBetween) 
    throws Exception {
      Iterator iter = inBetween.iterator();
      CutPoint virtualCut = new CutPoint(c,cutValue,
					 ((Instance)iter.next()).classValue());
      while ( iter.hasNext() ) {
	virtualCut = new CutPoint(virtualCut,cutValue,
				  ((Instance)iter.next()).classValue());
      } // end of while ()
      this.copyValues(virtualCut);
      if ( debug > 0 ) {
	System.out.println("*****************************");
	System.out.println("Real cut :\n" + this.toInfoString());      
	System.out.println("*****************************");	
      } // end of if ()
  }
  
  /*
    Constructor for virtual cuts only
   */

  private CutPoint(CutPoint c, double cutValue, double classVal) {
    attr = c.getAttribute();
    cut = cutValue;
    classValue = classVal;
    insts = c.getInstances();
    updateValues(c);
    if ( debug > 1 ) {    
      System.out.println("\n===========================");
      System.out.println("Previous cut :\n" + c.toInfoString());
      System.out.println("---------------------------");
      System.out.println("Virtual cut :\n" + this.toInfoString());
      System.out.println("===========================");
    } // end of if ()
  }

  private void copyValues(CutPoint virtualCut){
    attr = virtualCut.getAttribute();
    cut = virtualCut.getCut();
    classValue = virtualCut.getClassValue();
    discerned = virtualCut.getDiscerned();
    left = virtualCut.getLeftOfCut();
    right = virtualCut.getRightOfCut();
    leftModClass = virtualCut.getLeftModClass();
    rightModClass = virtualCut.getRightModClass();
    insts = virtualCut.getInstances();
  }
  
  /**
   * Get the value of debug.
   * @return value of debug.
   */
  public int getDebug() {
    return debug;
  }
  
  /**
   * Set the value of debug.
   * @param v  Value to assign to debug.
   */
  public void setDebug(int  v) {
    this.debug = v;
  }
  

  public Instances getInstances(){
    return insts;
  }

  public double getClassValue() {
    return classValue;
  }

  public int[] getLeftModClass() {
    return leftModClass;
  }

  public int[] getRightModClass() {
    return rightModClass;
  }

  public int getLeftModClass(int i) {
    return leftModClass[i];
  }

  public int getRightModClass(int i) {
    return rightModClass[i];
  }

  public int getLeftModDecision() {
    return leftModClass[(int)getClassValue()];
  }

  public int getRightModDecision() {
    return rightModClass[(int)getClassValue()];
  }

  public Attribute getAttribute() {
    return attr;
  }
  public double getCut() {
    return cut;
  }
  public int getDiscerned() {
    return discerned;
  }
  public int getLeftOfCut() {
    return left;
  }

  public RoughInstances getLeftOfCut(RoughInstances r) {
    RoughInstances res = new RoughInstances(r,0);
    for (int i=0;i<r.size();i++) {
      RoughInstance rInst = (RoughInstance)r.get(i);
      if (rInst.value(attr) < cut) {
	res.add(rInst);
      } // end of if ()
    } // end of for ()
    return res;
  }

  public int getRightOfCut() {
    return right;
  }

  public RoughInstances getRightOfCut(RoughInstances r) {
    RoughInstances res = new RoughInstances(r,0);
    for (int i=0;i<r.size();i++) {
      RoughInstance rInst = (RoughInstance)r.get(i);
      if (rInst.value(attr) >= cut) {
	res.add(rInst);
      } // end of if ()
    } // end of for ()
    return res;
  }

  public int getDiscerned(RoughInstances insts) {
    int [] leftModClass = new int[insts.numClasses()];
    int [] rightModClass = new int[insts.numClasses()];
    Arrays.fill(leftModClass,0);
    Arrays.fill(rightModClass,0);
    int left = 0;
    int right = 0;
    for (int i=0;i<insts.size();i++) {
      RoughInstance inst = insts.get(i);
      if ( inst.isMissing(attr) ) {
	continue;
      } // end of if ()
      if ( inst.value(attr) < cut ) {
	leftModClass[(int)inst.classValue()]++;
	left++;
      } // end of if ()
      else if ( inst.value(attr) > cut ) {
	rightModClass[(int)inst.classValue()]++;
	right++;
      } // end of if ()	
    } // end of for ()
    int sum = 0;
    for (int i=0;i<insts.numClasses();i++) {
      sum += leftModClass[i] * rightModClass[i];
    } // end of for ()
    return left*right - sum;
  }

  public int getDiscerned(SortedClasses classes) {
    int sum = 0;
    for (int i=0;i<classes.size();i++) {
      sum += getDiscerned(classes.getClass(i));
    } // end of for ()
    return sum;
  }

  public boolean leftOfCut(Instance i) {
    return i.value(attr) < cut;
  }

  public boolean equals(Object o) {
    CutPoint c = (CutPoint)o;
    return (attr.equals(c.getAttribute())) && (cut == c.getCut());
  }

  public String toString() {
    StringBuffer s = new StringBuffer();
    s.append("(" + attr.name() + " = " + cut + ")");
    return s.toString();
  }

  public String toInfoString() {
    StringBuffer s = new StringBuffer();
    s.append(this.toString() + "\n");
    s.append("ClassValue: " + this.getClassValue()+"\n");
    s.append("Number discerned: " + this.getDiscerned()+"\n");
    s.append("Left of cut in each class: ");
    for (int i=0;i<leftModClass.length;i++) {
      s.append(leftModClass[i]+(i<leftModClass.length-1?" ":"\n"));
    } // end of for ()
    s.append("Right of cut in each class: ");
    for (int i=0;i<rightModClass.length;i++) {
      s.append(rightModClass[i]+(i<rightModClass.length-1?" ":"\n"));
    } // end of for ()
    
    return s.toString();
  }

}

