/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MDCutPoints
 * Copyleft (C) 2002 Ola Leifler
 *
 */


package weka.roughset;

import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import weka.core.*;

/**
 *  <code>MDCutPoints</code> represents a set of cuts on numerical
 *  attributes, used in
 *  <code>weka.filters.DynamicReductDiscretize</code> and
 *  <code>weka.filters.RSDiscretize</code>.
 *
 * @author <a href="mailto:olale@aste1.ida.liu.se">Ola Leifler</a>
 * @version 1.0 */

public class MDCutPoints extends Object {

  private HashSet cutsVector;

  private Iterator iterator(){
    return cutsVector.iterator();
  }

  public MDCutPoints(Reduct r) {
    
  }

  /**
   * Creates a new <code>MDCutPoints</code> instance by sorting the
   * instances on each successive numerical attribute and creating
   * <code>SortedClasses</code>-objects each time, representing the
   * instances. Each class represents a set of objects that all have
   * the same value on the current numerical attribute. Those sets are
   * given as arguments for the constructors of each
   * <code>CutPoint</code> that is constructed.
   *
   * @param i an <code>Instances</code> value
   * @exception Exception if an error occurs */
  
  public MDCutPoints(Instances i) throws Exception { 
    cutsVector = new HashSet();
    Enumeration attrEnum = i.enumerateAttributes();
    CutPoint currentCut = null; 
    double max=0,min=0;
    Collection inBetween=null;
    Attributes attrs = new Attributes();
    while ( attrEnum.hasMoreElements() ) {
      Attribute attr = (Attribute)attrEnum.nextElement();
      if ( attr.isNumeric()  ) {
	attrs.clear();
	attrs.add(attr);
	SortedClasses s = new SortedClasses(i,attrs);
	for (int j=0;j<s.size();j++ ) {
	  max = s.getClass(j).get(0).value(attr);
	  if ( j==1 ) {
	    currentCut = new CutPoint(attr,(min+max)/2,inBetween,i);
	  } else if ( j>1 ) {
	    currentCut = new CutPoint(currentCut,(min+max)/2,inBetween);	    
	  } // end of if ()
	  if ( j > 0 ) {
	    addCut(currentCut);  	    
	  } // end of if ()
	  inBetween = s.getClass(j).getRoughInstances();
	  min=max;
	} // end of for ()
      } // end of if () 
    } // end of while ()
    
  }



  /**
   * Dummy constructor. Creates a new empty <code>MDCutPoints</code>
   * instance.
   * */
  
  public MDCutPoints() {
    cutsVector = new HashSet();
  }

  

  /**
   *  Add a new cutPoint to the set.
   *
   * @param cut a <code>CutPoint</code> value
   * @exception Exception if the argument is null.
   */
  
  public void addCut(CutPoint cut) throws Exception{
    if ( cut != null ) {
      cutsVector.add(cut);      
    } else {
      throw new Exception("Who did this?");
    } // end of else
    
  }

  public CutPoint getCut(int i) {
    Iterator iter=cutsVector.iterator();
    for (int j=0;iter.hasNext()&&j<i;j++,iter.next()) {} // end of for ()
    return (CutPoint)iter.next();
  }

  public boolean removeCut(CutPoint c) {
    return cutsVector.remove(c);
  }



  /**
   * Get the cut on attribute <b>a</b>that discerns the most pairs of
   * objects modulo their decision values. Discernibility is computed
   * by each <code>CutPoint</code>.
   *
   * @param a an <code>Attribute</code> value
   * @return a <code>CutPoint</code> value */
  
  public CutPoint getMaxCut(Attribute a) {
    CutPoint maxCut = null;
    for (Iterator i = iterator();i.hasNext();) {
      CutPoint next = (CutPoint)i.next();
      if ( !next.getAttribute().equals(a)) {
	continue;
      } // end of if ()
      if ( (maxCut == null) || 
	   (maxCut.getDiscerned() < next.getDiscerned()) ) {
	maxCut = next;
      } // end of if ()      
    } // end of for ()
    return maxCut;
  }


  /**
   * Get the cutpoint that discerns the most objects.
   *
   * @return a <code>CutPoint</code> value
   */
  
  public CutPoint getMaxCut(){
    Iterator i = iterator();
    CutPoint maxCut = null;
    while ( i.hasNext() ) {
      CutPoint next = (CutPoint)i.next();
      if ( (maxCut == null) || 
	   (maxCut.getDiscerned() < next.getDiscerned()) ) {
	maxCut = next;
      } // end of if ()
    } // end of while ()    
    return maxCut;
  }

  public CutPoint getMaxCut(SortedClasses classes) throws NullPointerException {
    Iterator i = iterator();
    CutPoint maxCut = null;
    while ( i.hasNext() ) {
      CutPoint next = (CutPoint)i.next();
      if ( (maxCut == null && next.getDiscerned(classes) > 0) || 
	   (maxCut != null && 
	    (maxCut.getDiscerned(classes) < next.getDiscerned(classes))) ) {
	maxCut = next;
      } // end of if ()
    } // end of while ()    
    if ( maxCut == null ) {
      throw new NullPointerException("No best cut from " + toString() 
				     + " on the classes\n" + classes.toString());
    } // end of if ()
    return maxCut;
  }

  

  /**
   * Is attribute <b>a</b> discretized?
   *
   * @param a an <code>Attribute</code> value
   * @return a <code>boolean</code> value
   */
  
  public boolean attributeDiscretized(Attribute a) {
    Iterator i = cutsVector.iterator();
    boolean res = false;
    while ( i.hasNext() ) {
      if ( a.equals(((CutPoint)i.next()).getAttribute()) ) {
	res = true;
	break;
      } // end of if ()
    } // end of while ()
    return res;
  }

  /**
   * Add all possible cuts for the nondiscretized attributes in the set
   * @param inst an <code>Instances</code> value
   */
  public void fillNonDiscretized(Instances inst) {
    try {
       
      Iterator iter = (new Attributes(inst)).iterator();
      double previous = Double.NaN, current;
      while ( iter.hasNext() ) {
	Attribute a = (Attribute)iter.next();	
	if ( a.isNumeric() && !attributeDiscretized(a) ) {
	  inst.sort(a);	  
	  for (int i=0;i<inst.numInstances();i++) {
	    current = inst.instance(i).value(a);
	    if ( i > 0 
		 && !inst.instance(i).isMissing(a) 
		 && !(previous == current) ) {
	      Collection c = new ArrayList();
	      c.add(inst.instance(i));
	      cutsVector.add(new CutPoint(a,previous+current/2,c,inst));
	    } // end of if ()
	    previous = current;
	  } // end of for ()
	  previous = Double.NaN;
	} // end of if ()
      } // end of while ()
  

    } catch (UnsupportedAttributeTypeException e) {
      System.err.println("Error in MDCutPoints:fillNonDiscretized: " + 
			 e.getMessage());
      e.printStackTrace();
    } // end of try-catch
    
  }


  /**
   *   <p>Create a new dataset where each cutpoint becomes a binary
   *   attribute. Encode "less than" as 0 and "greater than or equal"
   *   as 1 (see toAttributes() below). Then, insert the data from the
   *   old table into the new one. We will want to use this table for
   *   calculating a reduct, which will correspond to a set of cuts in
   *   the old table.</p>
   *
   *
   * @param insts an <code>Instances</code> value
   * @return an <code>Instances</code> value */
  public Instances binaryInstances(Instances insts) {

    /**
       (1) Create the attribute vector, where each cutpoint becomes a
       binary attribute. Encode "less than" as 0 and "greater than or
       equal" as 1 (see toAttributes() below) */


    /**
       (2) Create the new dataset
     */

    FastVector v = this.toReduct(insts).getAttributes().toFastVector();
    v.addElement(insts.classAttribute());

    Instances newInsts = new Instances(insts.relationName() + "-binary",v,
				       insts.numInstances());
    newInsts.setClassIndex(v.size()-1);

    /**
       (3) Create the instances in the new dataset
     */

    for (int i=0;i<insts.numInstances();i++) {
      Instance inst = insts.instance(i);
      int numAttrs = newInsts.numAttributes();
      double [] vals = new double[numAttrs];
      for (int j=0;j<numAttrs;j++) {
	if (j<numAttrs-1) {
	  vals[j] = (getCut(j).leftOfCut(inst)?0:1);	       	  
	} else {
	  vals[j] = inst.classValue();
	} // end of else
      } // end of for ()
      Instance newInst = null;
      if (inst instanceof SparseInstance) {
	newInst = new SparseInstance(inst.weight(), vals);
      } else {
	newInst = new Instance(inst.weight(), vals);
      }
      newInst.setDataset(newInsts);
      newInsts.add(newInst);
    } // end of for ()
    return newInsts;
  }



  /**
   * Translate the cuts to a <code>Reduct</code> which is simply a set
   * of attributes associated with a new table of binary values.
   *
   * @param insts an <code>Instances</code> value
   * @return a <code>Reduct</code> value */

  public Reduct toReduct(Instances insts){

    Attributes newAttrs = new Attributes();
    FastVector binaryValues = new FastVector(2);
    binaryValues.addElement("0");
    binaryValues.addElement("1");
    for (int i=0;i<size();i++) {
      newAttrs.add(new Attribute(getCut(i).toString(),binaryValues,i));
    } // end of for ()
    return new Reduct(this,newAttrs,insts);

  }



  /**
   * Translate a set of attributes into the corresponding set of cuts
   * by this method. Only matches the names of the attributes to the
   * <code>toString()</code>-values of the cutspoints in this set and
   * keeps those that are matched to an attribute name.
   *
   * @param a an <code>Attributes</code> value
   * @return a <code>MDCutPoints</code> value
   * @exception Exception if an error occurs */
  
  public MDCutPoints getSubset(Attributes a) throws Exception{
    MDCutPoints md = new MDCutPoints(); 
    for (Iterator cutsIter=iterator();cutsIter.hasNext(); ) {
      boolean exists=false;
      CutPoint c = (CutPoint)cutsIter.next();
      String s = c.toString();
      for (Iterator attrIter=a.iterator();attrIter.hasNext()&&!exists; ) {
	exists=((Attribute)attrIter.next()).name().equals(s);
      } // end of for ()
      if ( exists ) {
	md.addCut(c);
      } // end of if ()
    } // end of for ()
    return md;
  }

  
  /**
   * We use the representation of cutpoints that
   * <code>weka.filters.Discretizefilter</code> uses so that we can
   * steal the <code>convertInstance(Instance i)</code>-method from
   * there in our own classes that use this class as the
   * representation of cutpoints.
   *
   * @param insts an <code>Instances</code> value
   * @return a <code>double[][]</code> value */
  
  public double [][] toDoubleArray(Instances insts) {
    double [][] result = new double[insts.numAttributes()][];
    try {	
      for (int i=0;i<insts.numAttributes();i++) {
	Attribute a = insts.attribute(i);
	ArrayList valArray = new ArrayList();
	
	for (int j=0;j<cutsVector.size();j++) {
	  if ( getCut(j).getAttribute().equals(a) ) {
	    valArray.add(new Double(getCut(j).getCut()));
	  } // end of if ()
	} // end of for ()	
	if ( valArray.size() > 0 ) {
	  double [] vals = new double [valArray.size()];
	  for (int k=0;k<valArray.size();k++) { 
	    vals[k] = ((Double)valArray.get(k)).doubleValue();
	  } // end of for ()
	  Arrays.sort(vals);
	  result[i] = vals;
	} // end of if ()
	else {
	  result[i] = null;
	} // end of else
      } // end of for ()
    } catch (Exception e) {
      e.printStackTrace();
    } // end of try-catch
    return result;
  }

  public String toString() {
    StringBuffer s = new StringBuffer();
    for (Iterator i = cutsVector.iterator();i.hasNext();) {
      s.append(((CutPoint)i.next()).toString() + "\n");      
    } // end of for ()
    return s.toString();
  }



  /**
   * Outputs not only the attributes and values of the cuts, but also
   * discernibility-related information
   *
   * @return a <code>String</code> value */
  public String toInfoString() {
    StringBuffer s = new StringBuffer();
    /**
       Append something informative to s 
    */
    for (Iterator i=cutsVector.iterator();i.hasNext();) {
      s.append(((CutPoint)i.next()).toInfoString() + "\n");      
    } // end of for ()
    
    return s.toString();
  }
  

  public boolean equals(Object o) {
    return ((MDCutPoints)o).toString().equals(this.toString());
  }

  public int size() {
    return cutsVector.size();
  }

  /**
   * Method for testing the <code>MDCutPoints</code>
   *
   * @param args a <code>String[]</code> value
   */
  
  public static void main (String[] args) {
    try {
      Reader reader = null;
      if (args.length == 0) {
	reader = new BufferedReader(new InputStreamReader(System.in));
      } else {
	reader = new BufferedReader(new FileReader(args[0]));
      }
      Instances i = new Instances(reader);
      i.setClassIndex(i.numAttributes()-1);
      MDCutPoints c = new MDCutPoints(i);
      System.out.println("CutPoints on a sample dataset:\n" 
			 + c.toString() );
      System.out.println("Binary Instances:\n" 
			 + c.binaryInstances(i).toString());
      
    } catch (Exception e) {
      e.printStackTrace();
    } // end of try-catch


    
  } // end of main ()
  

}

