/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Reduct.java
 *    Copyright (C) 2002 Ola Leifler
 *
 */

package weka.roughset;

import java.io.*;
import java.util.*;
import weka.core.*;

public class Reduct extends Object {

  private Attributes attributes;
  private Instance instance;
  private Instances instances;
  private Reducts reductSet;
  private MDCutPoints cuts;

  public Reduct(MDCutPoints c,Attributes a,Instances i){
    cuts = c;
    attributes = a;
    instances = i;
    instance = null;
  }  

  public Reduct(Attributes a, Instance i) {
    instance = i;
    attributes = a;
    instances = null; 
  }

  public Reduct(Attributes a, Instance i,Instances insts) {
    instance = i;
    attributes = a;
    instances = insts; 
  }

  /**
   * Get the value of instance.
   * @return value of instance.
   */
  public Instance getInstance() {
    return instance;
  }
  
  /**
   * Set the value of instance.
   * @param v  Value to assign to instance.
   */
  public void setInstance(Instance  v) {
    this.instance = v;
  }

  /**
   * Get the value of cuts.
   * @return value of cuts.
   */
  public MDCutPoints getCuts() {
    return cuts;
  }
  
  /**
   * Set the value of cuts.
   * @param v  Value to assign to cuts.
   */
  public void setCuts(MDCutPoints  v) {
    this.cuts = v;
  }
  
  /**
   * Get the value of attributes.
   * @return value of attributes.
   */
  public Attributes getAttributes() {
    return attributes;
  }
  
  /**
   * Set the value of attributes.
   * @param v  Value to assign to attributes.
   */
  public void setAttributes(Attributes  v) {
    this.attributes = v;
  }
  
  
  /**
   * Get the value of instances.
   * @return value of instances.
   */
  public Instances getInstances() {
    return instances;
  }
  
  /**
   * Set the value of instances.
   * @param v  Value to assign to instances.
   */
  public void setInstances(Instances  v) {
    this.instances = v;
  }
  
  
  /**
   * Get the value of compareDataSets.
   * @return value of compareDataSets.
   */
  public boolean getCompareDataSets() {
    return (reductSet != null) && reductSet.getCompareDataSets();
  }
  
  /**
   * Set the value of compareDataSets.
   * @param v  Value to assign to compareDataSets.
   */
  public void setCompareDataSets(boolean  v) {
    reductSet.setCompareDataSets(v);
  }
    
  /**
   * Get the value of reductSet.
   * @return value of reductSet.
   */
  public Reducts getReductSet() {
    return reductSet;
  }
  
  /**
   * Set the value of reductSet.
   * @param v  Value to assign to reductSet.
   */
  public void setReductSet(Reducts  v) {
    this.reductSet = v;
  }
  
  public boolean equals(Object o) {
    boolean res = (o instanceof Reduct);
    if ( res ) {    
      Reduct r = (Reduct)o;
      res = !getCompareDataSets() || instances.equals(r.getInstances());
      if ( res && attributes != null) {
	res = attributes.equals(r.getAttributes()); 
      } // end of if ()
    } // end of if ()
    return res;
  }

  /**
     In the reduct set that this reduct is part of, how many other
     reducts have the same set of attributes but different reduct
     domains? Divide that number by the total number of subtables in
     the reduct set. Strangely
  */

  public double stability(){
    boolean b = getCompareDataSets();
    setCompareDataSets(false);
    Iterator iter = reductSet.iterator();
    double numEquals = 0; 
    while ( iter.hasNext() ) {
      Reduct r = (Reduct)iter.next();
      numEquals += (this.equals(r)?1:0);
    } // end of while ()
    setCompareDataSets(b);
    return numEquals/reductSet.getNumSubtables();
  }

  public boolean subsetOf(Reduct r) {
    return attributes.subsetOf(r.getAttributes());
  }

  /**
     Do the equivalence classes induced by the attributes in 'a' have
     the same decision values? In such a case we have a relative
     reduct, albeit not necessarily a proper one.
  */

  public static boolean isReduct(Attributes a,Instances i) {
    Attributes attrs = new Attributes(i);
    attrs.removeAll(a);
    return (new SortedClasses(i,a).bestAttribute(attrs) == null);
  }

  /**
     Can we remove an attribute from a reduct and still have a reduct?
     in such case it is not a proper reduct.
   */

  public static boolean isProperReduct(Attributes a,Instances i) {
    Reduct r = new Reduct(a,null,i);
    return Reduct.trimReduct(r).equals(r);
  }

  /**
     Algorithm 6 in some Efficient Algorithms for Rough Set methods
   */

  public static Reduct trimReduct(Reduct superReduct) {
    Instances instances = superReduct.getInstances();
    SortedClasses s = new SortedClasses(instances,
					superReduct.getAttributes());
    int m = RoughSet.positiveRegionSize(s);
    Iterator attrIter = superReduct.getAttributes().iterator();
    Attributes properReduct = new Attributes(superReduct.getAttributes());
    Attribute a = null;
    while ( attrIter.hasNext() ) {
      a = (Attribute)attrIter.next();
      Attributes red = new Attributes(properReduct); 
      red.remove(a);
      s = new SortedClasses(instances,red);
      int n = RoughSet.positiveRegionSize(s);
      if ( n == m ) {	
	properReduct.remove(a);
      } // end of if ()
    } // end of while ()
    Reduct res = new Reduct(properReduct,superReduct.getInstance(),instances);
    return res;
  }

  public String toString() {
    StringBuffer s = new StringBuffer();
    /** 
	Append something informative to s 
    */
    s.append(attributes.toString());
    return s.toString();
  }

  public String toInfoString(){
    StringBuffer s = new StringBuffer();
    /** 
	Append something informative to s 
    */
    s.append(this.toString()+"\n");
    for (int i=0;i<instances.numInstances();i++) {
      s.append(instances.instance(i) +"\n");
    } // end of for ()
    return s.toString();    
  } 

  public int hashCode(){
    return attributes.hashCode();
  }

  public int size(){
    return attributes.size();
  }

  public static void main (String[] args) {
     
  } // end of main ()
  

}
