/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*   RelativeReduct.java
 *   Copyright (C) 2002 Ola Leifler
 */

package weka.roughset;

import java.io.*;
import java.util.*;
import weka.core.*;

/**
 * Class for Relative Reducts, i.e. sets of attributes that maintain the
 * discernability function between instances. 
 *
 * @author <a href="mailto:olale@aste1.ida.liu.se">Ola Leifler</a>
 * @version 1.0
 */

public class RelativeReductComputation {

  static int debug;
  static int DEBUG_LEVELS = 3;
  private Instances instances;

  public RelativeReductComputation(Instances i){
    instances = i;
  }

  public RelativeReductComputation(Instances i,int d){
    instances = i;
    debug = d;
  }

  /**
   * Get the value of debug.
   * @return value of debug.
   */
  public int getDebug() {
    return debug;
  }
  
  /**
   * Set the value of debug.
   * @param v  Value to assign to debug.
   */
  public void setDebug(int  v) {
    this.debug = v;
  }



  /**
   * Returns the minimal set of attributes that discerns instance at
   * position <b>index</b> from other instances.
   *
   * @param index an <code>int</code> value
   * @return a <code>Reduct</code> value */
  public  Reduct minimalRelativeReduct(int index) {
    int numAttr = instances.numAttributes();
    Instance k = instances.instance(index);
    BitSet b = new BitSet(numAttr);
    b.set(0,numAttr-1,true);
    for (int i=0;i<instances.numInstances();i++) {
      Instance l = instances.instance(i);
      BitSet b1 = new BitSet(numAttr);
      if ( l.classValue() == k.classValue() ) {
	b1.set(0,numAttr-1,true);
      } // end of if ()
      else {      
	for (int j=0;j<numAttr;j++) {
	  b1.set(j,(j != l.classIndex() && l.value(j) != k.value(j)));
	} // end of for ()
	if ( debug >= 1 ) {
	  System.out.println("Different attributes between instances " 
			     + index + " and " + i + ": " + b1.toString());
	} // end of if ()
      } // end of else
      b.and(b1);
    } // end of for ()
    Attributes a = new Attributes();
    for (int i=0;i<numAttr;i++) {
      if ( b.get(i) ) {
	a.add(instances.attribute(i));
      } // end of if ()
    } // end of for ()
    return new Reduct(a,k,instances);
  }

  

  /**
   * Starting from the attributes in <b>old</b>, build a superreduct
   * that maintains the indiscernibility relation that is given by the
   * whole set of attributes by using a greedy algorithm that selects
   * the attribute that discerns the most objects as returned by the
   * <code>weka.roughset.SortedClasses.bestAttribute(Attributes
   * attrs)</code> method.
   *
   * @param old an <code>Attributes</code> value
   * @param instances an <code>Instances</code> value
   * @return a <code>Reduct</code> value */
  
  public static Reduct semiMinimalReduct(Attributes old, Instances instances) {
    Attributes reduct = new Attributes(old); 
    Attributes attrs = new Attributes(instances);
    Attribute a = new Attribute("Dummy");    
    while ( !( (a == null) || attrs.isEmpty() ) ) {
      SortedClasses s = new SortedClasses(instances,reduct);
      if ( debug >= 2 ) {	
	s.printClasses();
      } // end of if ()
      a = s.bestAttribute(attrs);
      if ( debug >= 2 ) {
	System.out.println("Best attribute: " 
			   + (a != null ? a.name() : "null"));
      } // end of if ()      
      if ( a != null ) {
	// A := A\{a}
	attrs.remove(a);
	reduct.add(a);	
      } // end of if ()
    } // end of while ()
    return new Reduct(reduct,null,instances);
  }



  /**
   * Works as <code>semiMinimalReduct</code>, only we use a
   * non-deterministic algorithm that uses
   * <code>weka.roughset.SortedClasses.randomBestAttribute(Attributes
   * attrs)</code> for selecting each successive attribute.
   *
   * @return a <code>Reduct</code> value */
  
  public Reduct randomReduct(){
    Attributes reduct = new Attributes();
    Attributes attrs = new Attributes(instances);
    Attribute a = new Attribute("Dummy");    
    while ( !((a == null) || attrs.isEmpty()) ) {
      SortedClasses s = new SortedClasses(instances,reduct);
      a = s.randomBestAttribute(attrs);
      if ( a != null ) {
	// A := A\{a}
	attrs.remove(a);
	reduct.add(a);	
      } // end of if ()
    } // end of while ()
    return new Reduct(reduct,null,instances);
  }

  /**
   * Obtaina a reduct by using the method described in <i>Genetic
   * algorithms in decomposition of classification problems, Jakub
   * Wrblewski</i> in <i>Rough Sets in Knowledge Discovery 2:
   * Applications, Case Studies and Software Systems, 1998</i>
   *
   * @param perm an <code>int[]</code> value
   * @return a <code>Reduct</code> value */
  
  public Reduct wroblewskiReduct(int[] perm){
    Attributes attrs = new Attributes(instances);
    if ( debug >= 2 ) {
      System.out.println("Chosen permutation:");
      for (int i=0;i<perm.length;i++) {
	System.out.print(perm[i] + (i<perm.length-1?", ":"\n"));
      } // end of for ()
    } // end of if ()
    

    Attributes rems = new Attributes();
    for (int i=0;i<perm.length;i++) {
      rems.add(attrs.get(perm[i]));
    } // end of for ()
    
    for (int i=0;i<rems.size();i++) {
      Attribute a = (Attribute)rems.get(i);
      attrs.remove(a);
      if (!Reduct.isProperReduct(attrs,instances) ) {
	attrs.add(a);
      } // end of if ()      
    } // end of for ()
    return new Reduct(attrs,null,instances);
  }

  /**
   * Obtains a set of reducts by using the method described in
    <i>Dynamic reducts as a tool for extracting laws from decisions
    tables, Jan G. Bazan and Andrzej Skowron and Piotr Synak 1994</i>
   *
   * @param choices an <code>int</code> value
   * @param confidence a <code>double</code> value
   * @param maxError a <code>double</code> value
   * @param stability a <code>double</code> value
   * @param numReducts an <code>int</code> value
   * @return a <code>Reducts</code> value
   * @exception Exception if an error occurs */
  
  public Reducts dynamicReducts(int choices, double confidence, 
				double maxError, double stability, 
				int numReducts) 
    throws Exception{
    
    SubtableChooser s = new SubtableChooser(instances,choices,
					    confidence,maxError);    
    Reducts reducts = new Reducts();
    reducts.setCompareDataSets(true);

    /* Extract all reducts from the subtables using a naive
       implementation. Then simply count those reducts that are more
       frequent than <stability> dictates */

    while ( s.hasNext() && numReducts > 0) {      
      RelativeReductComputation rel = 
	new RelativeReductComputation(s.next());
      rel.setDebug(debug);
      if ( rel.getDebug() >= 2 ) {
	System.out.println("Generating relative reducts in subset:" );
      } // end of if ()      
      Permutations p = new Permutations(instances.numAttributes()-1);
      while ( p.hasNext() && numReducts > 0) {
	Reduct r = rel.wroblewskiReduct(p.next());
	if (r != null ) {
	  if ( debug >= 2 ) {
	    System.out.println("Adding reduct " + r.toString());
	  } // end of if ()
	  int size = reducts.size();
	  reducts.add(Reduct.trimReduct(r));
	  numReducts -= (reducts.size() > size ? 1:0);
	} // end of if ()
      } // end of while ()
      if ( debug >= 2 ) {
	System.out.println("Reducts in last subset:\n" 
			   + reducts.toString());
      } // end of if ()
      
    } // end of while ()
    reducts.filter(stability);
    return reducts;
  }

  /*
    There should really be a method here that extracts all the proper
    reducts, but we'll make do with only some of them for the sake of
    efficiency
  */

  public static Reducts someProperReducts(Reduct superReduct,int numIterations) 
    throws Exception{
    Reducts properReducts = new Reducts();
    properReducts.setCompareDataSets(false);
    Instances i = superReduct.getInstances();
    Permutations p = new Permutations(superReduct.size());
    int oldSize;
    MDCutPoints cuts;
    Reduct next;
    Attributes a;
    while (p.hasNext()&&numIterations-->0) {
      a = new Attributes(superReduct.getAttributes(),p.next());

	/**
	   We have a set of cutpoints for one table that is supposed
	   to be the set of attributes for another table. In order to
	   be able to keep this convention we use this ugly method of
	   setting the cuts int the new reduct to the set
	   corresponding to the attributes
	 */
      
      next = Reduct.trimReduct(new Reduct(a,null,i));
      cuts = superReduct.getCuts();
      if ( cuts != null ) {
	cuts = cuts.getSubset(next.getAttributes());
      } // end of if ()
      next.setCuts(cuts);
      if ( !properReducts.containsSubsetOf(next ) ) {
	oldSize = properReducts.size();
	properReducts.add(next);	
	//        numReducts -= (properReducts.size()>oldSize?1:0);
      } // end of if ()
    }
    return properReducts;
  }

  /**
     
     This simple heuristic takes a random reduct and reduces it into a
     set of proper reducts 

  */

  public Reducts someProperReducts(int numIterations) throws Exception{
    Reducts properReducts = new Reducts();
    while ( numIterations-- > 0 ) {
      Reduct red = randomReduct();
      if ( !properReducts.containsSubsetOf(red) ) {
	if ( !Reduct.isProperReduct(red.getAttributes(),red.getInstances()) ) {
	  Reducts reds = RelativeReductComputation.
	    someProperReducts(red,numIterations);
	  for (int i=0;i<reds.size();i++) {
	    properReducts.add(reds.get(i));
	  } // end of for ()	  
	} else {
	  properReducts.add(red);
	} // end of else
      }
    } // end of while ()
    return properReducts;
  }


  public static void main (String[] args) {
    try {
      Reader reader = null;
      if (args.length == 0) {
	reader = new BufferedReader(new InputStreamReader(System.in));
      } else {
	reader = new BufferedReader(new FileReader(args[0]));
      }
      Instances i = new Instances(reader);
      i.setClassIndex(i.numAttributes()-1);
      RelativeReductComputation.debug = 1;
      RelativeReductComputation r = new RelativeReductComputation(i,2);

//        for (Permutations p = new Permutations(i.numAttributes()-1);
//  	   p.hasNext(); ) {
//  	System.out.println("Wroblewskis reduct (trimmed):\n" 
//  			   + r.trimReduct(r.wroblewskiReduct(p.next())));
//        } // end of for ()
      
      // Reducts dynReducts = r.dynamicReducts(5, 0.1, 0.15, 0.6, 4);
      // System.out.println("\nDynamic reducts:\n" + dynReducts.toString());
      Reducts reds = r.someProperReducts(4);
      System.out.println("Set of some proper reducts:\n" + reds.toString());
   } catch (Exception e) {
      System.err.println(e.getMessage());
      e.printStackTrace();
    } // end of try-catch
  } // end of main ()
  

}

