/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    RoughInstances.java
 *    Copyright (C) 2002 Ola Leifler
 *
 */

package weka.roughset;

import java.util.*;
import weka.core.*;
import java.io.*;

/* This class should implement something that resembles a List, whether
   on it's own or via some superclass */

public class RoughInstances extends Instances  {

  private ArrayList roughInstances;
  private Attributes activeAttributes;
  private AttributeValuePairs pairs;
  private boolean compareAll;
  private boolean compareDecision = false;
  private boolean missingIsWildcard = true;

  private int numSupporting(Instances insts, AttributeValuePair pair){
    Enumeration iter = insts.enumerateInstances();
    Attribute a = pair.getAttribute();
    double val = pair.getValue();
    int numSupporting = 0;
    while ( iter.hasMoreElements() ) {
      Instance i = (Instance)iter.nextElement();
      if ( i.value(a) ==  val ) {
	numSupporting++;
      } // end of if ()
    } // end of while ()
    return numSupporting;
  }

  private void initPairs(){
    pairs = new AttributeValuePairs();    
    if ( size() > 0  ) {
      Iterator iter = iterator();
      while ( iter.hasNext() ) {
	RoughInstance inst = (RoughInstance)iter.next();
	for (int i=0;i<inst.numAttributes();i++) {
	  Attribute attr = inst.attribute(i);
	  if ( !attr.equals(inst.classAttribute()) && 
	       !inst.isMissing(attr) ) {
	    pairs.addPair(attr,inst.value(attr));	  	    
	  } // end of if ()
	} // end of for ()
      } // end of while ()    
    } // end of if ()    
  }
    
  public RoughInstances(RoughInstances rInst) {
    super(rInst,0);
    roughInstances = rInst.copy();
    Iterator iter = roughInstances.iterator();
    while ( iter.hasNext() ) {
      RoughInstance r = (RoughInstance)iter.next();
      r.setRoughDataSet(this);
    } // end of while ()
    pairs = rInst.getAllPairs();
    activeAttributes = rInst.getActiveAttributes();
  }

  public RoughInstances(Instances inst) {
    super(inst,0);
    activeAttributes = new Attributes(inst);
    Enumeration e = inst.enumerateInstances();
    roughInstances = new ArrayList(inst.numInstances());
    while ( e.hasMoreElements() ) {
      roughInstances.add(new RoughInstance((Instance)e.nextElement(),
					   this));
    } // end of while ()
  }

  public RoughInstances(RoughInstances insts, int capacity) {
    super(insts,0);
    activeAttributes = insts.getActiveAttributes();
    roughInstances = new ArrayList(capacity);
    pairs = null;
  }

  public RoughInstances(Instances inst, Attributes a) {
    super(inst,0);
    activeAttributes = a;
    Enumeration e = inst.enumerateInstances();
    roughInstances = new ArrayList(inst.numInstances());
    while ( e.hasMoreElements() ) {
      roughInstances.add(new RoughInstance((Instance)e.nextElement(),
					   this));
    } // end of while ()
    initPairs();
  }

  public RoughInstances(Instances inst, AttributeValuePairs p) {
    this(inst);
    retainMatching(p);
  }

  public ArrayList copy() {
    RoughInstance[] dest = new RoughInstance[roughInstances.size()];
    System.arraycopy(roughInstances.toArray(dest),
		     0,dest,0,roughInstances.size());
    return (new ArrayList(Arrays.asList(dest)));
  }


  public void add(RoughInstance r) {
    r.setRoughDataSet(this);
    roughInstances.add(r);
  }

  public void remove(int i){
    roughInstances.remove(i);
  }

  public void remove(RoughInstance r) {
     roughInstances.remove(r);
  }

  public void addAll(RoughInstances r) {
    roughInstances.addAll((new RoughInstances(r)).getRoughInstances());
  }

  /* Change the attribute set that describes the rough instances */

  public void setActiveAttributes(Attributes a) {
    activeAttributes = a;
  }

  public Attributes getActiveAttributes() {
    return (compareAll ? new Attributes(this) : activeAttributes);
  }
  
  /**
   * Get the value of compareDecision.
   * @return value of compareDecision.
   */
  public boolean isCompareDecision() {
    return compareDecision;
  }
  
  /**
   * Set the value of compareDecision.
   * @param v  Value to assign to compareDecision.
   */
  public void setCompareDecision(boolean  v) {
    this.compareDecision = v;
  }

  /**
   * Get the value of missingIsWildcard.
   * @return value of missingIsWildcard.
   */
  public boolean missingIsWildcard() {
    return missingIsWildcard;
  }
  
  /**
   * Set the value of missingIsWildcard.
   * @param v  Value to assign to missingIsWildcard.
   */
  public void setMissingIsWildcard(boolean  v) {
    this.missingIsWildcard = v;
  }
  

  public void sort() {
    Collections.sort(roughInstances);
  }

  public void sort(Attributes a) {
    Attributes attr = activeAttributes;
    setActiveAttributes(a);
    this.sort();
    setActiveAttributes(attr);
  }
  
  /**
   * Get the value of CompareAll.
   * @return value of CompareAll.
   */

  public boolean isCompareAll() {
    return compareAll;
  }
  
  /**
   * Set the value of CompareAll.
   * @param v  Value to assign to CompareAll.
   */
  public void setCompareAll(boolean  v) {
    this.compareAll = v;
  }
  

  public Collection getRoughInstances() {
    return roughInstances;
  }

  public AttributeValuePair getBestPair() {
    return pairs.getBestPair(this);
  }

  public AttributeValuePairs getAllPairs(){
    initPairs();
    return pairs;
  }
  
  public int numSupporting(AttributeValuePair pair){
    Iterator iter = iterator();
    Attribute a = pair.getAttribute();
    double val = pair.getValue();
    int numSupporting = 0;
    while ( iter.hasNext() ) {
      RoughInstance i = (RoughInstance)iter.next();
      if ( i.value(a) ==  val ) {
	numSupporting++;
      } // end of if ()
    } // end of while ()
    return numSupporting;
  }

  /**
   * remove duplicates 
   * @exception Exception if too many inconsistencies were detected
   */
  public void compact() throws Exception {
    boolean b = isCompareAll();
    boolean d = isCompareDecision();
    setCompareDecision(false);
    setCompareAll(true);
    HashSet h = new HashSet();
    int numBefore = size();
    int numInconsistent=0;
    /**
       Set this value to .05 if working with reasonably sized datasets
     */

    double fraction = .5;

    for (Iterator i=iterator();i.hasNext();) {
      RoughInstance r = (RoughInstance)i.next();
      if ( !h.add(r) ) {
	i.remove();
	numInconsistent += (negativeMatch(r)!=null?1:0);
      }
    } // end of for ()
    if ( numInconsistent > numBefore*fraction ) {
      throw new Exception((double)numInconsistent*100/(double)numBefore 
			  + " % of the instances discarded, " 
			  + "please use GeneralInstances instead");
    } // end of if ()
    setCompareDecision(d);
    setCompareAll(b);
  }

  /* Extract a rough instance that is equal *modulo the attributes in the
     rough set*, but has different decision value */

  public RoughInstance negativeMatch(RoughInstance r1) {
    Iterator i = this.iterator();
    boolean d = isCompareDecision();
    setCompareDecision(false);
    while ( i.hasNext() ) {
      RoughInstance r2 = (RoughInstance)i.next();
      if ( r1.equals(r2) && r1.classValue() != r2.classValue() ) {
	return r2;
      } // end of if ()
    } // end of while ()
    setCompareDecision(d);
    return null;
  }

    

  public RoughInstances retainMatching(AttributeValuePair p){
    Iterator i = this.iterator();
    RoughInstances res = new RoughInstances(this,0);
    while ( i.hasNext() ) {
      RoughInstance r = (RoughInstance)i.next();
      if ( !p.matches(r) ) {
	i.remove();
	res.add(r);
      } // end of if ()
    } // end of while ()
    return res;
  }

  public RoughInstances retainMatching(AttributeValuePairs p){
    Iterator i = this.iterator();
    RoughInstances res = new RoughInstances(this,0);
    while ( i.hasNext() ) {
      RoughInstance r = (RoughInstance)i.next();
      if ( !p.matches(r) ) {
	i.remove();
	res.add(r);
      } // end of if ()
    } // end of while ()
    return res;
  }

  public RoughInstances retainMatching(LocalCovering l){
    Iterator i = l.iterator();
    RoughInstances res = new RoughInstances(this,0);
    while ( i.hasNext() ) {
      res.addAll(this.retainMatching((AttributeValuePairs)i.next()));
    } // end of while ()
    return res;
  }

  public void removeMatching(AttributeValuePair p){
    Iterator i = this.iterator();
    while ( i.hasNext() ) {
      RoughInstance r = (RoughInstance)i.next();
      if ( p.matches(r) ) {
	i.remove();
      } // end of if ()
    } // end of while ()
  }

  public void removeMatching(AttributeValuePairs p){
    Iterator i = this.iterator();
    while ( i.hasNext() ) {
      RoughInstance r = (RoughInstance)i.next();
      if ( p.matches(r) ) {
	i.remove();
      } // end of if ()
    } // end of while ()
  }

  public void removeMatching(LocalCovering p){
    Iterator i = this.iterator();
    while ( i.hasNext() ) {
      RoughInstance r = (RoughInstance)i.next();
      if ( p.matches(r) ) {
	i.remove();
      } // end of if ()
    } // end of while ()
  }

  public boolean equals(Object o) {

    /* A very generous comparison that checks if the Rough Instances,
       however if in different classes, are the same in both sets */

    boolean comp = true;
    if ( !(o instanceof RoughInstances) ) {
      return false;
    } // end of if ()
    RoughInstances r = (RoughInstances)o;    
    boolean b1,b2;
    b1 = this.isCompareAll();
    b2 = r.isCompareAll();
    this.setCompareAll(true);
    r.setCompareAll(true);
    Iterator iter1 = this.iterator();
    Iterator iter2 = r.iterator();
    while ( iter1.hasNext() && iter2.hasNext() ) {      
      RoughInstance i1 = (RoughInstance)iter1.next();
      RoughInstance i2 = (RoughInstance)iter2.next();
      if ( i1.compareTo(i2) != 0 ) {
	comp = false;
	break;
      } // end of if ()
    } // end of while ()
    if ( iter1.hasNext() || iter2.hasNext() ) {
      comp = false;
    } // end of if ()
    this.setCompareAll(b1);
    r.setCompareAll(b2);
    return comp;
  }

  public boolean moreSupported(AttributeValuePair p1,
			       AttributeValuePair p2) {
    return 
      (p1 != null) && 
      (p2 != null) && 
      (numSupporting(p1) > numSupporting(p2));
  }

  public Iterator iterator() {
    return roughInstances.iterator();
  }

  public Attributes getAllAttributes() { 
    return new Attributes(this);
  }

  public int size() {
    return roughInstances.size();
  }

  public boolean isEmpty() {
    return roughInstances.isEmpty();
  }

  public RoughInstance get(int i) {
    return (RoughInstance)roughInstances.get(i);
  }

  public boolean contains(RoughInstance r) {
    return roughInstances.contains(r);
  }


  public boolean containsAll(RoughInstances r) {
    HashMap h = new HashMap(r.size());
    Integer one = new Integer(1);
    for (int i=0;i<r.size();i++) {
      h.put(r.get(i),one);
    } // end of for ()
    for (int j=0;j<this.size();j++) {
      RoughInstance rInst = this.get(j);
      if ( h.containsKey(rInst) ) {
	int val = ((Integer)h.get(rInst)).intValue();
	h.put(rInst,new Integer(val+1));
      } // end of if ()
    } // end of for ()
    Collection c = h.values();
    return !c.contains(one);
  }

  public boolean containsSome(RoughInstances r) {
    HashMap h = new HashMap(r.size());
    Integer one = new Integer(1);
    boolean res = false;
    for (int i=0;i<r.size();i++) {
      h.put(r.get(i),one);
    } // end of for ()
    for (int j=0;j<this.size();j++) {
      RoughInstance rInst = this.get(j);
      if ( h.containsKey(rInst) ) {
	res = true;
	break;
      } // end of if ()
    } // end of for ()
    return res;
  }


  public String toString(){
    StringBuffer s = new StringBuffer();
    Iterator i = iterator();
    while ( i.hasNext() ) {
      s.append(((RoughInstance)i.next()).toString());
    } // end of while ()
    return s.toString();
  }

  public void printInfo() {
    System.out.println(this.toString());
  }
  
  public static void main (String[] args) {
    try {
      Reader r = null;
      if (args.length == 0) {
	r = new BufferedReader(new InputStreamReader(System.in));
      } else {
	r = new BufferedReader(new FileReader(args[0]));
      }
      Instances i = new Instances(r);
      i.setClassIndex(4);
      Attributes a = new Attributes(i);
      AttributeValuePair p = new AttributeValuePair(i.classAttribute(), 1);

      RoughInstances rough = new RoughInstances(i,a);
      rough.removeMatching(p);
      System.out.println("\n=========================================\n");
      System.out.println(rough.size() + " Rough Instances with " + 
			 i.classAttribute().name() + " != " + 1 + ": " );
      rough.printInfo();

      rough = new RoughInstances(i,a);
      rough.retainMatching(p);
      System.out.println("\n=========================================\n");
      System.out.println(rough.size() + " Rough Instances with " + 
			 i.classAttribute().name() + " == " + 1 + ": " );
      rough.printInfo();

      rough = new RoughInstances(i,a);
      System.out.println("Best attribute-value pair:" + 
			 rough.getBestPair().toString());
      a = new Attributes();
      a.add(i.attribute(0));
      rough = new RoughInstances(i,a);
      Attributes b = new Attributes();
      b.add(i.attribute(1));
      rough.setActiveAttributes(b);
      System.out.println("Sorting on " + i.attribute(1).name());
      rough.sort();
      System.out.println("Active attributes in the Rough Set: " 
			 + rough.getActiveAttributes().toString());
      rough.printInfo();
    } catch (Exception e) {
      System.err.println(e.getMessage());
      e.printStackTrace();
    } // end of try-catch

  } // end of main ()
  
  
}



