/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * RoughSet
 * Copyleft (C) 2002 Ola Leifler
 *
 */

package weka.roughset;

import java.io.*;
import java.util.*;
import weka.core.*;

/**
 * A number of static methods that implement basic RoughSet functionalities 
 *
 * @author <a href="mailto:olale@aste1.ida.liu.se">Ola Leifler</a>
 * @version 1.1
 */

public class RoughSet {



  /**
   * Have two instances the same value of the attributes in
   * <b>attributes</b>?
   *
   * @param i1 an <code>Instance</code>
   * @param i2 an <code>Instance</code>
   * @param attributes an <code>Attributes</code>
   * @return a <code>boolean</code> */
  
  public static boolean sameClass(Instance i1, Instance i2, 
				  Attributes attributes) {
    Iterator attrIter = attributes.iterator();
    while ( attrIter.hasNext() ) {
      Attribute a = (Attribute)attrIter.next();
      if ( i1.value(a) != i2.value(a) ) {
	return false;
      } // end of if ()
    } // end of while ()
    return true;
  }


  /**
   * Return true if all elements of <b>rInsts</b> have the same
   * decision value.
   *
   * @param rInsts a <code>RoughInstances</code> value
   * @return a <code>boolean</code> value */
  
  public static boolean sameDecision(RoughInstances rInsts){
    double firstDecision;
    boolean equal = true;
    Iterator iter = rInsts.iterator();
    if ( iter.hasNext() ) {
      firstDecision = ((RoughInstance)iter.next()).classValue();
      while ( iter.hasNext() && equal ) {
	equal = (firstDecision == ((RoughInstance)iter.next()).classValue());
      } // end of while ()
    } // end of if ()
    return equal;
    
  }

  /**
   * Returns true if <b>set</b> contains all elements of
   * <b>subSet</b>. Equality is tested modulo the active attributes
   * (@see weka.roughset.RoughInstances)
   *
   * @param subSet a <code>RoughInstances</code> value
   * @param set a <code>RoughInstances</code> value
   * @return a <code>boolean</code> value */
  
  public static boolean subsetOf(RoughInstances subSet, RoughInstances set){
    return set.containsAll(subSet);
  }

  /**
   * Returns true if at least one element of <b>set2</b> is in <b>set1</b>
   *
   * @param set1 a <code>RoughInstances</code> value
   * @param set2 a <code>RoughInstances</code> value
   * @return a <code>boolean</code> value
   */
  
  public static boolean nonEmptyIntersection(RoughInstances set1, 
					     RoughInstances set2){

      boolean b1,b2;
      b1 = set1.isCompareAll();
      b2 = set2.isCompareAll();
      boolean res = set1.containsSome(set2);
      set1.setCompareAll(b1);
      set2.setCompareAll(b2);
      return res;
  }  

  /**
   * Compute the lower approximation to a set <b>i</b> by using the
   * method described in <i>Some efficient algorithms for Rough Set
   * methods</i>, Hung Son Nguyen and Sinh Hoa Nguyen, in
   * <i>Information Processing and Management of Uncertainty in
   * Knowledge-Based Systems</i> 1996.
   * @param s a <code>SortedClasses</code> value
   * @param i a <code>RoughInstances</code> value
   * @return a <code>RoughInstances</code> value */
  
  public static RoughInstances lowerApproximation(SortedClasses s, 
						  RoughInstances i){
    RoughInstances approx = new RoughInstances(i,0);
    for (Iterator iter = s.iterator();iter.hasNext(); ) {
      RoughInstances r = (RoughInstances)iter.next();
      if ( RoughSet.subsetOf(r,i) ) {
	approx.addAll(r);
      } // end of if ()      
    } // end of for ()
    return approx;
  }

  public static RoughInstances lowerApproximation(Instances instances,
						  AttributeValuePair d) {
    
    /*
      This should really be a call to
      lowerApproximation(SortedClasses,RoughInstances).. */

    AttributeValuePairs newClass = new AttributeValuePairs();
    newClass.addPair(d);
    RoughInstances certainInstances = new RoughInstances(instances);
    RoughInstances res = new RoughInstances(certainInstances,0);
    for (int i=0;i<instances.numInstances();i++) {
      certainInstances = new RoughInstances(instances);
      RoughInstance r = (RoughInstance)certainInstances.get(i);
      if ( r.classValue() == d.getValue() ) {
	certainInstances.remove(r);
	RoughInstance duplicate = certainInstances.negativeMatch(r);
	if ( duplicate == null ) {
	  res.add(r);
	} // end of if ()
      } // end of if ()
    } // end of for ()
    return res;
  }


  public static RoughInstances upperApproximation(SortedClasses s, 
						  RoughInstances i){
    RoughInstances approx = new RoughInstances(i,0);
    Iterator e = s.iterator();
    while ( e.hasNext() ) {
      RoughInstances r = (RoughInstances)e.next();
      if ( RoughSet.nonEmptyIntersection(r,i) ) {
	  approx.addAll(r);
      } // end of if ()
    } // end of while ()    
    return approx;
  }

  public static RoughInstances upperApproximation(Instances instances,
						  AttributeValuePair decision){
    /*
      This should really be a call to
      upperApproximation(SortedClasses,RoughInstances).. */

    AttributeValuePairs newClass = new AttributeValuePairs();
    newClass.addPair(decision);
    return new RoughInstances(instances,newClass);
  }

  /** Calculate the minimal size of a subtable to use for 
     stability calculation */
  
  public static int minSubtableSize(Instances i,double accuracy,
			 double error,Reducts r,int maxIterations) 
    throws Exception{
    double fraction = .5;
    int numSubtables = 10;
    boolean cont = true;
    int subtableSize = 0;
    while ( cont && maxIterations > 0) {
      
      subtableSize = (int)(i.numInstances()*fraction);
      
      double ratio = 0;
      SubtableChooser chooser = new SubtableChooser(i,subtableSize,
						    numSubtables);
      while ( chooser.hasNext() ) {
	ratio += ((double)r.numProperReducts(i))/((double)r.size());
      } // end of while ()
      ratio /= numSubtables;
      if ( ratio > accuracy + error ) {
	fraction -= fraction/2;
      } else if ( ratio < accuracy - error ) {
	fraction += fraction/2;
      } // end of if ()
      else {
	cont=false;
      } // end of else
      maxIterations--;
    } // end of while ()
    return subtableSize;
    
  }

  public static int positiveRegionSize(SortedClasses s) {
    int size = 0;
    for (Iterator iter = s.iterator();iter.hasNext();) {
      RoughInstances r = (RoughInstances)iter.next();
      size += (RoughSet.sameDecision(r)?r.size():0);
    } // end of for ()
    return size;
  }



  public static void main (String[] args) {
     
  } // end of main ()
  

}

