/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Rules.java
 *    Copyright (C) 2002 Ola Leifler
 *
 */

package weka.roughset;

import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Enumeration;
import weka.core.Instances;
import weka.core.Instance;

public class Rules {

  private HashMap ruleSets;
  private Instances instances;

  public Instances getInstances() {
    return instances;
  }

  public Rules(Instances i){
    instances = i;
    ruleSets = new HashMap();
  }

  public void addRule(Rule r){
    this.addRule(r,r.getDecision());
  }

  public void addRule(Rule r,AttributeValuePair key){
    if ( !ruleSets.containsKey(key) ) {
      ruleSets.put(key,new RuleSet(this,key));
    } // end of if ()
    ((RuleSet)ruleSets.get(key)).addRule(r);
  }

  public void addRules(LocalCovering c, AttributeValuePair d) {
    if (!ruleSets.containsKey(d) ) {
      ruleSets.put(d,new RuleSet(this,c,d));
    } // end of if ()
    ((RuleSet)ruleSets.get(d)).addRules(c);
  }

  public void addRules(Rules r) {
    for (Iterator i = r.iterator(); i.hasNext(); ) {
      RuleSet subSet = (RuleSet)i.next();
      AttributeValuePair p = subSet.getDecision();
      if ( !ruleSets.containsKey(p) ) {
	ruleSets.put(p,subSet);
      } // end of if ()
      ((RuleSet)ruleSets.get(p)).addAll(subSet);	
    } // end of while ()
  }

  /* Negotiate a rule according to one of several schemes */

  public double getDecision (Instance i, String m) throws Exception {
    double maxStrength = Double.NEGATIVE_INFINITY;
    double strength = Double.NEGATIVE_INFINITY;
    double decision = Double.NaN;
    
    for (Iterator iter = ruleSets.values().iterator(); iter.hasNext(); ) {
      RuleSet ruleSet = (RuleSet)iter.next();
      if ( m.equals("s")  ) {
	strength = ruleSet.simpleStrength(i);
      } // end of if ()
      else if ( m.equals("m")  ) {
	strength = ruleSet.maximalStrength(i);
      } // end of if ()
      else if ( m.equals("r")  ) {
	strength = ruleSet.globalStrength(i);
      } // end of if ()
      else if ( m.equals("d") ) {
	strength = ruleSet.stabilityStrength(i);
      } // end of if ()
      else if ( m.equals("p")  ) {
	strength = ruleSet.globalStabilityStrength(i);
      } // end of if ()
      else if ( m.equals("c")  ) {
	strength = ruleSet.conditionalStrength(i);
      } // end of if ()
      else if ( m.equals("a")  ) {
	strength = ruleSet.partialSupportStrength(i);
      } // end of if ()
      else {
	throw(new Exception("Invalid parameter '" + m + "'"));
      } // end of else
      
      if ( strength > maxStrength ) {
	maxStrength = strength;
	decision = ruleSet.getDecisionValue();
      } // end of if ()
    } // end of while ()
    if ( Double.isNaN(decision) ) {
      decision = Instance.missingValue();
    } // end of if ()
    return decision;
  }

  public void approximateRules(double consistency) {
    for (Iterator i = iterator();i.hasNext(); ) {
      ((RuleSet)i.next()).approximateRules(consistency);            
    } // end of for ()
  }

  public void reduceRuleSets(double consistency) {
    for (Iterator i = iterator();i.hasNext();) {
      ((RuleSet)i.next()).reduceRuleSet(consistency);            
    } // end of for ()
  }

  public Iterator iterator() {
    return ruleSets.values().iterator();
  }

  public int size() {
    return ruleSets.size();
  }

  /* The LEM2 algorithm, for certain and uncertain decision classes. */

  public void addRules(Instances instances, boolean certain, int debug) 
    throws Exception {
    AttributeValuePairs decisionClasses = 
      new AttributeValuePairs(instances);    
    RoughInstances b;
    RoughInstances g;
    LocalCovering conditions;
    AttributeValuePair newClass;
    AttributeValuePairs condition;
    AttributeValuePairs allRelevant;
    AttributeValuePair bestPair = new AttributeValuePair();

    // O(n log_2 n)

    // O (|d|)
    for (Iterator iter = decisionClasses.iterator();iter.hasNext(); ) {

      /* Generate Possible/Certain Rules */
      newClass = (AttributeValuePair)iter.next();
      if ( debug >= 2 ) {
	System.out.println("Generating " + 
			   (certain ? "certain" : "possible ") + 
			   " rules for " + 
			   newClass.toString());	
      } // end of if ()

      // O(n)
      b = (certain ? RoughSet.lowerApproximation(instances,newClass) : 
	RoughSet.upperApproximation(instances,newClass));
      g = new RoughInstances(b);

      /**
	 We do not consider inconsistent data yet. Throws an exception
	 if it doesn't work..  
      */

      b.compact();
      g.compact();
      
      conditions = new LocalCovering(instances);

      // O(n)
      while ( !g.isEmpty() ) {
	condition = new AttributeValuePairs();

	// O(n * m)
	allRelevant = g.getAllPairs();

	// O(m)
	while ( (condition.isEmpty() || 
		 
		 !RoughSet.subsetOf(new RoughInstances(instances,
						       condition),
				    b))
		&& (condition.size() < (instances.numAttributes()-1))) {
	  /* Just as LEM2, we ignore attribute priorities */	  
	  
	  // O (n * m)
	  bestPair = allRelevant.getBestPair(g);
	  if ( bestPair == null ) {
	    break;
	  } // end of if ()
	  condition.addPair(bestPair);	    
	  
	  // O (n)
	  g.retainMatching(bestPair);
	  if ( debug >= 3 ) {
	    System.out.println("Added " + bestPair.toString() + " and kept " +
			       g.size() + " instances with " + 
			       (RoughSet.sameDecision(g) ? 
				"same decision value" : 
				"different decision values"));
	  } // end of if ()
	  
	  // O (n * m)
	  allRelevant = g.getAllPairs();
	  // O (n * m)
	  allRelevant.removePairs(condition);
	} // end of while ()

	Iterator i = condition.iterator();

	/* Remove unnecessary pairs */

	// O (n * m)
	while ( i.hasNext() ) {
	  AttributeValuePair p = (AttributeValuePair)i.next();
	  AttributeValuePairs complex = new AttributeValuePairs(condition);
	  complex.removePair(p);
	  if ( RoughSet.subsetOf(new RoughInstances(instances,
						     complex),b) ) {
	    i.remove();
	  } // end of if ()
	} // end of while ()
	conditions.addPairs(condition);
	g = new RoughInstances(b);

	// O (n)
	g.removeMatching(conditions);
	if ( debug >= 3 ) {
	  System.out.println("Current conditions for " + 
			     newClass.toString() + " are " + 
			     condition.ruleString());
	} // end of if ()
      } // end of while ()
      if ( debug >= 2 ) {
	System.out.println("Conditions before pruning for " + 
			   newClass.toString() + " =\n" + 
			   conditions.toString());
	
      } // end of if ()

      // O (n^2 * m)
      conditions.trimToFit(b);     
      this.addRules(conditions,newClass);
      if ( debug >= 2 ) {
	System.out.println("Final conditions for " + 
			   newClass.toString() + " =\n" + 
			   conditions.toString());
	System.out.println("New rule set:\n" + this.toString());
	
      } // end of if ()
    } // end of for ()
  }

  /**
     Dynamic reduct rule generation
   */

  public void addRules(Reducts dynReducts) {
    
    /**
       Each reduct is a unique pair (Attributes x Instances), which
       means that two reducts may contain the same attributes but have
       different reduct domains (set of instances for which they are reducts).

       This may be an awkward way of identifying reducts
       theoretically, but algorithmically it works ok for me.

    */
    
    for (Iterator i = dynReducts.iterator();i.hasNext(); ) {
      Reduct red = (Reduct)i.next();
      Rules r = new Rules(red.getInstances());
      for (Enumeration e = red.getInstances().enumerateInstances();
	   e.hasMoreElements();) {
	Instance inst=(Instance)e.nextElement();
	if ( !inst.classIsMissing() ) {
	  r.addRule(new Rule(inst,red),new AttributeValuePair(inst));	  
	} // end of if ()
      } // end of for ()
      this.addRules(r);
    } // end of for ()    
  }

  public String toString() {
    StringBuffer s = new StringBuffer();
    for (Iterator i = iterator();i.hasNext(); ) {
      s.append(((RuleSet)i.next()).toString());
      s.append("\n");
    } // end of while ()
    return s.toString();
  }
}
