/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*   SortedClasses.java
 *   Copyright (C) 2002 Ola Leifler
 */

package weka.roughset;

import java.io.*;
import java.util.*;
import weka.core.*;

/**
 * Describe class <code>SortedClasses</code> here.
 *
 * @author <a href="mailto:olale@aste1.ida.liu.se">Ola Leifler</a>
 * @version 1.0
 */
public class SortedClasses {

  private ArrayList classes;

  public SortedClasses(Instances i, Attributes attribs) {
    this(new RoughInstances(i,attribs),attribs);
  }

  public SortedClasses(RoughInstances r, Attributes attribs) {
    Attributes attr = r.getActiveAttributes();
    classes = new ArrayList();
    r.setActiveAttributes(attribs);
    r.sort();
    Iterator iter = r.iterator();
    RoughInstances lastClass = new RoughInstances(r,0);
    while ( iter.hasNext() ) {
      RoughInstance inst = (RoughInstance)iter.next();
      if ( !lastClass.contains(inst) ) {
	/* Since the elements are sorted according to their
	   equivalence classes, we only need one pass to distinguish
	   the elements from each other (modulo the attributes in the
	   reduct) Every time we encounter an object that is not int
	   the last equivalence class we add another class and put it
	   int there. */
	if ( !lastClass.isEmpty() ) {
	  classes.add(lastClass);
	} // end of if ()
	lastClass = new RoughInstances(r,0);
      }
      lastClass.add(new RoughInstance(inst,lastClass));
    } // end of while ()
    if ( lastClass.size() > 0 ) {
      classes.add(lastClass);      
    } // end of if ()
    r.setActiveAttributes(attr);
  }

  public int numDiscerned(Attribute a){
    Attributes attr = new Attributes();
    attr.add(a);
    int sum = 0;
    for (Iterator iter = classes.iterator();iter.hasNext(); ) {
      RoughInstances equivClass = (RoughInstances)iter.next();
      SortedClasses s = new SortedClasses(equivClass,attr);
      int sizes = 0;
      
      for (Iterator sIter = s.iterator();sIter.hasNext(); ) {
	RoughInstances r = (RoughInstances)sIter.next();
	sizes += Math.pow(r.size(),2);	
      } // end of for ()
      sum += (int)(Math.pow(equivClass.size(),2) 
		   - sizes)/2;
    } // end of for ()
    return sum;
  }


  public Attribute bestAttribute(Attributes a){
    Attributes attr = new Attributes(a);
    Iterator iter = a.iterator(); 
    int max = 0;
    Attribute best = null;
    Attribute next;
    Attributes attributes = (!classes.isEmpty() ? 
			     ((RoughInstances)classes.get(0)).
			     getActiveAttributes() : null);
    attr.removeAll(attributes);
    iter = attr.iterator();
    while ( iter.hasNext() ) {
      next =  (Attribute)iter.next();
      int numDiscerned = this.numDiscerned(next);
      if ( numDiscerned > max ) {
	max = numDiscerned;
	best = next;
      } // end of if ()
    } // end of while ()
    return best;
  }

  public Attribute randomBestAttribute(Attributes a){
    Attributes attr = new Attributes(a);
    Iterator iter = a.iterator(); 
    Hashtable hashTable = new Hashtable();
    ArrayList keyValues = new ArrayList();
    Attribute best = null;
    Attribute nextAttribute;
    Attributes attributes = (!classes.isEmpty() ? 
			     ((RoughInstances)classes.get(0)).
			     getActiveAttributes() : null);
    attr.removeAll(attributes);
    iter = attr.iterator();
    while ( iter.hasNext() ) {
      nextAttribute =  (Attribute)iter.next();
      Integer nextValue = new Integer(this.numDiscerned(nextAttribute));
      hashTable.put(nextValue,nextAttribute);
      keyValues.add(nextValue);
    } // end of while ()
    int sum = 0;
    Enumeration keyEnum = hashTable.keys();
    while ( keyEnum.hasMoreElements() ) {
      sum += ((Integer)keyEnum.nextElement()).intValue();
    } // end of while ()
    long randKey = Math.round(sum * Math.random());
    Collections.sort(keyValues);
    Iterator keyIter = keyValues.iterator();
    /* Since the keys are sorted, we need only traverse the ArrayList
       once and pick whatever Attribute that has a weight that happens
       to come closest to the randKey */
    long intVal = 0;
    while ( keyIter.hasNext() ) {
      Integer i = (Integer)keyIter.next();
      intVal += i.intValue();
      if ( (intVal >= randKey) && (i.intValue() > 0) ) {
	best = (Attribute)hashTable.get(i);
	break;
      } // end of if ()
    } // end of while ()    
    return best;
  }

  public Iterator iterator() {
    return classes.iterator();
  }

  public boolean isEmpty(){
    return classes.isEmpty();
  }

  public int size() {
    return classes.size();
  }

  public void compact() throws Exception {
    for (Iterator iter = iterator();iter.hasNext(); ) {
      ((RoughInstances)iter.next()).compact();
    } // end of for ()
  }
    
  public boolean sameDecisions(){
    boolean sameDecision = true;
    for (int i=0;i<classes.size()&&sameDecision;i++) {
      sameDecision = RoughSet.sameDecision((RoughInstances)classes.get(i));
    } // end of for ()
    return sameDecision;
  }

  public RoughInstances getClass(int i) {
    return (RoughInstances)classes.get(i);
  }

  public void splitClasses(CutPoint cut) {
    Iterator i = iterator();
    ArrayList newClasses = new ArrayList();
    while (i.hasNext()) {
      RoughInstances r = (RoughInstances)i.next();
      if ( RoughSet.sameDecision(r) ) {
	i.remove();
	continue;
      } // end of if ()
      RoughInstances leftOfCut = cut.getLeftOfCut(r);
      RoughInstances rightOfCut = cut.getRightOfCut(r);	    
      if ( (leftOfCut.size() > 0) && 
	   (rightOfCut.size() > 0) ) {
	i.remove();
	newClasses.add(leftOfCut);
	newClasses.add(rightOfCut);
      } // end of if ()
    } // end of while ()
    if ( !newClasses.isEmpty() )  {
      classes.addAll(newClasses);	    
    } // end of if ()
  }
  
  public void printClasses() {
    System.out.println(this.toString());
  }

  public String toString(){
    StringBuffer s = new StringBuffer();
    RoughInstances r = (RoughInstances)classes.get(0);
    int numAttr = r.numAttributes();
    Iterator attrIter = r.getActiveAttributes().iterator();
    s.append("\nEquivalence classes modulo : ");
    while ( attrIter.hasNext() ) {
      s.append(((Attribute)attrIter.next()).name() + " ");
    } // end of while ()
    s.append("\n");
    int index = 0;
    for (Iterator classIter = classes.iterator();classIter.hasNext();) {
      s.append("Class " + index++ + "\n" 
	       + ((RoughInstances)classIter.next()).toString());      
    } // end of for ()
    return s.toString();
  }

  public static void main (String[] args) {
    try {
      Reader r = null;
      if (args.length == 0) {
	r = new BufferedReader(new InputStreamReader(System.in));
      } else {
	r = new BufferedReader(new FileReader(args[0]));
      }
      Instances i = new Instances(r);
      i.setClassIndex(i.numAttributes()-1);
      Attributes a = new Attributes();
      a.add(i.attribute(0));
      SortedClasses s = new SortedClasses(i,a);
      System.out.println("Do the classes " + s.toString() 
			 + "have the same decision? " + s.sameDecisions());
      System.out.println("Is there a best attribute? " 
			 + (s.bestAttribute(new Attributes(i))==null?
			    "no":"yes, " 
			    + s.bestAttribute(new Attributes(i)).name()));
      System.out.println("The classes before compacting:\n" + s.toString());
      s.compact();
      System.out.println("and after compacting:\n" + s.toString());
      int m = RoughSet.positiveRegionSize(s);
      a.add(i.attribute(1));
      System.out.println("Positive region with one attribute: " + m 
			 + ".\nWe add another attribute and get: " 
			 + RoughSet.positiveRegionSize(new SortedClasses(i,a)));
      
    } catch (Exception e) {
      System.err.println(e.getMessage());
    } // end of try-catch


  } // end of main ()
  

}


